/*********************************************************************************
** TESfaith.c  (for Morrowind Master and Plugin files).
**********************************************************************************
** The stuff that moves Mountains!
** This utility is meant for TES Master or Plugin files created
** for Morrowind: Elder Scrolls III. It is able to move, copy or delete
** entire landscape cells including the objects contained within these
** cells to entirely different locations; this is sometimes necessary if
** a landscape conflict occurs between two plugins.
**
** See the associated Readme file for information on how to use this
** utility.
**
** Paul Halliday (September 2003).
**
** License: Consult the GNU license; you're free to redistribute and modify
**          this code as much as you please, just give me some recognition
**          name somewhere if you modify or re-use the code. :-)
**          Coded on Linux, compiled for Windows (Using Visual C++, now the
**          free Cygwin compiler from Cygnus).
**          p.s. You will need a standard working getopt library to
**               to compile this code (Not all Windows compilers have this).
**
** Changes:
**	Version 0.95: Correctly jumps over unknown FRMR, SCDT and DODT fields.
**		      Added TA, CA and DA (work on all ext cells) commands.
**  	Version 0.94: Third Release (15-January-2003).
** 		  Added functionality to modify Scripts and Dialog commands.
**                Added -n (only save the changes to new plugin) option.
**                Added ability to pickup command line arguments from Rules file.
**		  Corrected negative float->integer cell calculations for
**		  determining which Interior->Exterior doors should be moved.
**  	Version 0.93: Second Release (14-December-2003).
**		  Added functionality to modify Interior->Exterior door co-ordinates.
**  	Version 0.92: Original Release (26-September-2003).
************************************************************************************/

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

/* UNIX */
#include <unistd.h>

/* Windoze */
//#include <io.h>
//#include <fstream.h>
#include "getopt.h"

/***********************
** Function Prototypes.
**********************/

int  ProcessCELLData(char *r, int size);
int  ProcessFRMRData(char *r, int record_size);
int  ProcessDODTData(char *r, int record_size);
int  ProcessLANDData(char *r, int size);
int  ProcessSCPTData(char *r, int *record_size);
int  ProcessINFOData(char *r, int *record_size);
int  ModifyScriptLine(char *line);
void ReadRulesCLArgs(char *filename, int *argc, char *argv[]);
void ReadRules(char *filename, char *cell_record);
int  ParseRule(char *line);
int  bytes_to_int(char b1, char b2, char b3, char b4);
void Log(char *info);
void WriteCellChangeList(int old_x, int old_y, int trans_x, int trans_y);
int  CheckCellChangeList(int old_x, int old_y, int *trans_x, int *trans_y);
void LogCell(void);

enum { EXTERIOR, INTERIOR, TRUE, FALSE };

struct cell_data {
	int size;
	char name[1024];
	char region_name[1024];
	int current_x;
	int current_y;
	int new_x;
	int new_y;
	int type;
	int save;
	int copy;
} cell;

/***************************************************************
* The rules array is made global, only because it saves sending
* pointers between the ReadRules and ParseRules functions.
**************************************************************/

char rule[5][1024];

int verbose_mode = 0,	       /* Command line option: Option to produce an extensive log file.             */
    list_cell_mode = 0,	       /* Command line option: Option to produce a file listing the plugin's cells. */
    show_rules_mode = 0,       /* Command line option: Option to display verbose form of rules then exit.   */
    modify_script_mode = 1,    /* Command line option: Option to modify scripts if necessary: Default on.   */
    generate_specific_esp = 0, /* Command line option: Option to create a new ESP only containing changes.  */
    pass = 0;		       /* Current file pass. Tesfaith runs through the ESP/ESM in 2 passes.         */

char log_message[512];

/***************************
* Just some running totals.
**************************/

int total_cells = 0,		    /* Total CELL records found in the file. */
    total_land = 0,		    /* Total LAND records found in the file. */
    total_records_changed = 0,      /* LAND or CELL records changed/copied.  
*/
    total_objects = 0,		    /* Total objects found in the file.      */
    total_objects_changed = 0,      /* Total objects changed/copied.   	     
*/
    total_scripts = 0,		    /* Total SCPT records found in the file. */
    total_scripts_changed = 0,	    /* Total SCPT records modified.          
*/
    total_dialogs = 0,		    /* Total INFO scripts found in the file. */
    total_dialogs_changed = 0;	    /* Total INFO records modified.          
*/

#define TESFAITH_RULES_FILE 	     	"tfaith.cfg"
#define TESFAITH_LOG_FILE  	     	"tfthlog.txt"
#define TESFAITH_CHANGE_LIST_FILE    	"tfcells.tmp"
#define TESFAITH_CELL_LIST_FILE      	"cell_log.txt"
#define TESFAITH_NEW_ESP_FILE	     	 "TFOUT.ESP"
#define TESFAITH_TMP_SCRIPT_OUT_FILE 	 "tfscpt0.txt"
#define TESFAITH_TMP_NEW_SCRIPT_OUT_FILE "tfscpt1.txt"

/*********************************************************************
** 1. main(): The Main routine processes command line options,
**            opens the specified TES file as input, the new esp
**            file as output and reads each record in to memory from the
**            TES file, until no more records are found. If the current
**            record is a CELL or LAND record, then it is sent to
**            ProcessCELLData() or ProcessLANDData() respectively, to
**            extract the CELL (and object) co-ordinates or Landscape
**            co-ordinates, accordingly.
**            Before reading the next record, the current record is
**            saved to the new file (any changes to the data in memory
**            are carried out in ProcessCELLData(), ProcessFRMRData()
**            or ProcessLANDData().
*********************************************************************/

int main(int argc, char *argv[])
{
	int size,	/* Size of current record.             */
	    new_size;   /* Size of modified version of record. */

	char c;		/* For command-line getopts args.  */
	char s[40];	/* For storing the 16-byte header. */
	char *or;	/* Pointer to the Original Record. */
	char *nr;	/* Pointer to the New Record.      */

	FILE *fpin, /* Input File Stream (original ESP/ESM).      */
	     *fpout;/* Output FileStream (TESFAITH_NEW_ESP_FILE). */

	/*
	 * Invalid args. Try and get them from the Rules file if not on the command line.
	 */

	if (argc < 2) {
		fprintf(stderr, "No command line options specified: Searching for options in %s\n",
			TESFAITH_RULES_FILE);
		ReadRulesCLArgs(TESFAITH_RULES_FILE, &argc, argv);
	}

	if (argc < 2) {
		fprintf(stderr, "Usage: %s [-r -v -l -n -d] original.esp\n\n"
		   "\t-v : Verbose mode: Log file (%s) contains lots of output\n"
		   "\t-l : List all cells in ESP file to %s\n"
		   "\t-n : Create a new ESP containing only copied land, cells and objects.\n"
		   "\t-r : Display verbose version of the Rules in %s, then exit.\n"
		   "\t-d : Don't modify Scripts or Dialog Commands.\n"
		   "\t-V : Display current Version, then exit.\n",
		   argv[0], TESFAITH_LOG_FILE ,
		   TESFAITH_CELL_LIST_FILE, TESFAITH_RULES_FILE);
		sleep(5);
		exit(1);
	}

	/***********************************
	 * Parse the command line arguments.
	 **********************************/
	while ((c = getopt(argc, argv, "dlnrvV")) != EOF) {
		switch (c) {
			case 'l':
				unlink(TESFAITH_CELL_LIST_FILE);
				list_cell_mode = 1;
				break;
			case 'r':
				show_rules_mode = 1;
				break;
			case 'd':
				modify_script_mode = 0;
				break;
			case 'n':
				generate_specific_esp = 1;
				break;
			case 'v':
				verbose_mode = 1;
				break;
			case 'V':
				printf("TESfaith Version 0.95: (Sep-2003 -> 19-Feb-2004): Paul Halliday 
(Ocean_Lightwave@yahoo.co.uk)\n");
				exit(0);
			default:
				fprintf(stderr, "Usage: %s [-r -v -d -l -n] original.esp\n\n"
				   "\t-v : Verbose mode: Log file (%s) contains lots of output\n"
				   "\t-l : List all cells in ESP file to %s\n"
				   "\t-n : Create a new ESP containing only copied land, cells and objects.\n"
				   "\t-r : Display verbose version of the Rules in %s, then exit.\n"
				   "\t-d : Don't modify Scripts or Dialog Commands.\n"
				   "\t-V : Display current Version, then exit.\n",
				   argv[0], TESFAITH_LOG_FILE ,
				   TESFAITH_CELL_LIST_FILE, TESFAITH_RULES_FILE);
				exit(1);
		}
	}

	/*
	 * Initialize variables:
	 */

	cell.name[0] = '\0';
	cell.region_name[0] = '\0';
	cell.current_x = 0;
	cell.current_y = 0;
	cell.new_x = 0;
	cell.new_y = 0;
	cell.size = 0;
	cell.type = INTERIOR;
	cell.save = TRUE;
	cell.copy = FALSE;

	/* Delete the current log file. */

	unlink(TESFAITH_LOG_FILE);
	unlink(TESFAITH_CHANGE_LIST_FILE);

	if (!modify_script_mode)
		Log("The \"-d\" option has been specified: Scripts and Dialog Commands will not be modified.\n");


	/*********************************************************
	 * Parse the user Rules; the first time this is run, it'll
	 * just output a verbose interpretation of the rules.
	 * If we're in (-r) show rules mode then exit cleanly once
	 * the rules have been parsed.
	 *********************************************************/

	cell.type = EXTERIOR;
	ReadRules(TESFAITH_RULES_FILE, NULL);
	cell.type = INTERIOR;
	ReadRules(TESFAITH_RULES_FILE, NULL);
	if (show_rules_mode)
		exit(0);
		Log("--------------------------------------------"
		    "--------------------------------------------\n\n");

	for (pass = 1; pass <= 2; pass++) {

		/*
		 * Open argv[1] (input file) for reading.
		 */


		if ((fpin = fopen(argv[argc-1], "rb")) == NULL) {
			fprintf(stderr, "Unable to open %s for reading: %s\n",
				argv[argc-1], strerror(errno));
			exit(1);
		}

		/*
		 * Open argv[2] (output file) for writing.
		 */

		if (!list_cell_mode) {
			if ((fpout = fopen(TESFAITH_NEW_ESP_FILE, "wb")) == NULL) {
				fprintf(stderr, "Unable to open %s for writing: %s\n",
					argv[2], strerror(errno));
				exit(1);
			}
		}

		if (pass < 2 ) {
			printf("Step 1/2: Generating a list of cells that should be modifed in %s ...\n",
				argv[argc-1]);
		} else if (list_cell_mode) {
			printf("\nWriting a list of cells contained within %s to %s.\n",
				argv[argc-1], TESFAITH_CELL_LIST_FILE);
		} else {
			printf("\n\nStep 2/2: Processing data from \"%s\" and creating \"%s\":\n",
				argv[argc-1], TESFAITH_NEW_ESP_FILE);
		}

		/* TES records follows the format of TYPE (4 bytes) then
		 * the record length (4 bytes) - which we put in to the variable, size.
		 */

		while (fread(s, 1, 16, fpin) > 0) {

			cell.size = 0;
			cell.current_x = 0;
			cell.current_y = 0;
			cell.new_x = 0;
			cell.new_y = 0;
			cell.type = INTERIOR;
			cell.save = TRUE;
			cell.copy = FALSE;

			size = bytes_to_int(s[4], s[5], s[6], s[7]) + 16;
			cell.size = size;
			new_size = size;

			/* Create some memory space to store the full orignal record
			 * and an additional copy (nr) which may be modified.
			 * A cheap hack here adds 2048 bytes to the memory allocated
			 * to the new record. This is to cope with a compiler problem
			 * (with Cygwin when compiling a static binary) with reallocing
			 * more space (causes problems with free(or)) later.
			 * A Morrowind script can now grow by 2K without causing any
			 * problems - e.g. That's several hundred position updates in one script.
			 */

			if ((or = (void *) malloc(size)) == NULL
			 || (nr = (void *) malloc(size+2048)) == NULL) {
				fprintf(stderr, "Unable to allocate %d bytes of"
					" memory to store TES file record: %s\n",
					2*size, strerror(errno));
				exit(1);
			}

			fseek(fpin, -16, SEEK_CUR);

			if (fread(or, 1, size, fpin) < size) {
				fprintf(stderr, "Unable to read entire marker record"
					" (%d bytes) from %s into memory: %s\n",
					size, argv[argc-1], strerror(errno));
				exit(1);
			}

			/**********************************************************
			 * Copy the record to nr (if the record is modified then
			 * the modified version will be stored here; this helps
			 * keep both versions in case the user has asked for a copy
			 * of this cell to be stored somewhere else.
			 *********************************************************/

			memcpy(nr, or, size);

			if (verbose_mode && pass == 2) {
				sprintf(log_message, "New Record: %c%c%c%c: (%d bytes)",
					s[0], s[1], s[2], s[3], size);
				Log(log_message);
			}

			/******************************************************
			 ** If it's a CELL or a LAND record, then hand it on to
			 ** a procedure that will handle the format, including
			 ** determining if any modifications should be made.
			 *****************************************************/

			if (strncmp(s, "CELL", 4) == 0) {
				putchar(s[0]);
				fflush(stdout);
				cell.name[0] = '\0';
				cell.region_name[0] = '\0';
				total_cells++;
				ProcessCELLData(nr + 16, size);
			} else if (strncmp(s, "LAND", 4) == 0) {
				putchar(s[0]);
				fflush(stdout);
				total_land++;
				if (pass == 2) {
					ProcessLANDData(nr + 16, size);
				}
			} else if (strncmp(s, "SCPT", 4) == 0) {
				putchar(s[0]);
				fflush(stdout);
				if (pass == 2) {
					total_scripts++;
					if (modify_script_mode)	{
						ProcessSCPTData(nr, &new_size);
					} else if (verbose_mode) {
						Log("\n");
					}
				}
			} else if (strncmp(s, "INFO", 4) == 0) {
				putchar(s[0]);
				fflush(stdout);
				if (pass == 2) {
					total_dialogs++;
					if (modify_script_mode) {
						ProcessINFOData(nr, &new_size);
					} else if (verbose_mode) {
						Log("\n");
					}
				}
			} else {
				if (pass < 2 ) {
					putchar('-'); // Leave a trail of '.'s across screen ...
					fflush(stdout);
				} else {
					putchar('.');
					fflush(stdout);
				}
				if (verbose_mode && pass == 2) {
					sprintf(log_message, " - ignoring record (not CELL, LAND, SCPT or INFO)\n");
					Log(log_message);
				}
			}

			/************************************************************
			 * If this record is marked as save, then save the original
			 * unmodified record (*or) to the new file.
			 * If generate_specific_esp is set, only save the TES3 record.
			 ***********************************************************/

			if (pass == 2 && !list_cell_mode && cell.save == TRUE
			    && (!generate_specific_esp || strncmp(or, "TES3", 4) == 0)) {
				
				/* 
				 * If we're generating a fresh ESP of just the changes, make sure
				 * that byte 0x1C (28) is set to 0. ESM files have it set to 1.
				 */
				if (generate_specific_esp && strncmp(or, "TES3", 4) == 0) {
					or[28] = 0;
				}
				
				if (verbose_mode) {
					sprintf(log_message, "\t[Saving Original Record: %d bytes]\n", size);
					Log(log_message);
				}
				if (fwrite(or, 1, size, fpout) < size) {
					fprintf(stderr, "Unable to write new record data "
						" to %s: %s\n", argv[2], strerror(errno));
					exit(1);
				}
			}

			/**********************************************************
			 * If this record is marked as copy, then save the new
			 * modified record (*nr) to the new file.
			 *********************************************************/

			if (pass == 2 && !list_cell_mode && cell.copy == TRUE
			    && (!generate_specific_esp || strncmp(nr, "CELL", 4) == 0 || strncmp(nr, "LAND", 4) == 0)) {
				if (verbose_mode) {
					sprintf(log_message, "\t[Saving New (modified) Record: %d bytes]\n", new_size);
					Log(log_message);
				}
				if (fwrite(nr, 1, new_size, fpout) < new_size) {
					fprintf(stderr, "Unable to write new record data "
						" (%d bytes) to %s: %s\n", new_size, argv[2], strerror(errno));
					exit(1);
				}
			}
			free(or);
			free(nr);
		}
		fclose(fpin);

		if (list_cell_mode) {
			pass = 2;
			break;
		} else {
			fclose(fpout);
		}
	}


	/******************************
	 * Dump out the running totals.
	 *****************************/

	if (list_cell_mode) {
		sprintf(log_message, "\nTESFaith has finished. The list of cells has been written to %s.\n"
		"---------------------------------------------\n"
		"Total CELL records found:                %d\n"
		"---------------------------------------------\n",
		TESFAITH_CELL_LIST_FILE, total_cells);
		Log(log_message);
		printf(log_message);
	} else {
		sprintf(log_message, "\n\nTESFaith has finished. The new data has been written to %s\n\n"
		"\tTotal CELL records found:                     %d\n"
		"\tTotal LAND records found:                     %d\n"
		"\tTotal CELL/LAND records modified or copied:   %d\n"
		"\tTotal objects modified/found:                 %d/%d\n"
		"\tTotal Script Commands modified/Scripts found: %d/%d\n"
		"\tTotal Dialog entries  modified/found:         %d/%d\n",
		TESFAITH_NEW_ESP_FILE, total_cells, total_land, total_records_changed, total_objects_changed,
		total_objects, total_scripts_changed, total_scripts, total_dialogs_changed, total_dialogs);
		Log(log_message);
		printf(log_message);
	}

	unlink(TESFAITH_CHANGE_LIST_FILE);
	unlink(TESFAITH_TMP_SCRIPT_OUT_FILE);
	unlink(TESFAITH_TMP_NEW_SCRIPT_OUT_FILE);

	return 0;
}

/***************************************************************
** 2. ProcessCELLData:
** 	Processes a CELL record, extracting name, region and any
**	FRMR (objects in this cell) records (if they exist).
***************************************************************
** Format:
** CELL (4) + Length (12)
** NAME (4) + Length (4) + NameAsString (size given by Length)
** DATA (4) + Unknown (16).
** RGNN (4) + Length (4) + RegionAsString (size given by Length)
***************************************************************/

int ProcessCELLData(char *r, int size)
{
	int  nsize;
	int  pos = 0,
	     xypos = 0;

	/**********************************
	 * NAME of the cell (8+Name bytes).
	 *********************************/
	nsize = (int) r[4];
	strncpy(cell.name, r + 8, nsize);
	pos = 8 + nsize;

	/* The DATA section (4+16 = 20 bytes)
	 * Bytes 12-19 contain two integers that represent the (x, y)
	 * co-ordinates of this cell in the world grid.
	 * The variable xypos is a useful reference for this position,
	 * should we need to change these values later.
	 */
	xypos = pos + 12;
	cell.current_x = bytes_to_int(r[pos+12], r[pos+13], r[pos+14], r[pos+15]);
	cell.current_y = bytes_to_int(r[pos+16], r[pos+17], r[pos+18], r[pos+19]);
	cell.new_x = cell.current_x;
	cell.new_y = cell.current_y;

	pos += 20;
	/**********************************
	 * RGNN Region data (8+Name bytes).
	 *********************************/

	if (strncmp("RGNN", r + pos, 4) == 0 || cell.name[0] == '\0') {
		cell.type = EXTERIOR;
		if (strncmp("RGNN", r + pos, 4) == 0) {
			nsize = (int) r[pos+4];
			strncpy(cell.region_name, r + pos + 8, nsize);
			pos += 8 + nsize;
		}
		if (list_cell_mode) {
			LogCell();
			return 0;
		}
		if (verbose_mode && pass > 1) {
			Log(" [EXTERIOR] ");
			if (cell.name[0] == '\0') {
				Log("(Cell has no name) ");
			} else {
				sprintf(log_message, " \"%s\" ", cell.name);
				Log(log_message);
			}
			if (cell.region_name[0] == '\0') {
				Log("(No Associated Region) ");
			} else {
				sprintf(log_message, " \"%s\" ", cell.name);
				Log(log_message);
			}
			sprintf(log_message, " (%d, %d) ",
				cell.current_x, cell.current_y);
			Log(log_message);
		}
		ReadRules(TESFAITH_RULES_FILE, NULL);
		if (cell.current_x != cell.new_x || cell.current_y != cell.new_y) {

			/* If a cell is being moved, make a note of the cell co-ordinates.
			 * This is the entire purpose of the first pass. The list of cells
			 * to be changed can then be backreferenced by other records, such
			 * as Interior->Exterior Teleport doors, scripts, NPCs and Dialogue commands.
			 */

			if (pass == 1 && !list_cell_mode) {
				if (cell.save == FALSE) {
					WriteCellChangeList(cell.current_x, cell.current_y, cell.new_x - cell.current_x, cell.new_y - cell.current_y);
				}
				return 0;
			}

			if (!verbose_mode && pass > 1) {
				Log(" [EXTERIOR] ");
				if (cell.name[0] == '\0') {
					Log("(Cell has no name) ");
				} else {
					sprintf(log_message, " \"%s\" ", cell.name);
					Log(log_message);
				}
				if (cell.region_name[0] == '\0') {
					Log("(No Associated Region) ");
				} else {
					sprintf(log_message, " \"%s\" ", cell.region_name);
					Log(log_message);
				}
				sprintf(log_message, " (%d, %d) ",
					cell.current_x, cell.current_y);
				Log(log_message);
			}
			memcpy(&r[xypos], &cell.new_x, 4);
			memcpy(&r[xypos+4], &cell.new_y, 4);
			total_records_changed++;
			sprintf(log_message, "- MODIFIED to (%d, %d).\n",
				bytes_to_int(r[xypos], r[xypos+1], r[xypos+2], r[xypos+3]),
				bytes_to_int(r[xypos+4], r[xypos+5], r[xypos+6], r[xypos+7]));
			Log(log_message);
		} else if (verbose_mode && pass == 2)
			Log("\n");
	} else {   /* INTERIOR cell.  */
		cell.type = INTERIOR;
		if (verbose_mode && pass == 2) {
			sprintf(log_message, " [INTERIOR] \"%s\" ",
				cell.name);
			Log(log_message);
		}
		ReadRules(TESFAITH_RULES_FILE, r);
		if (verbose_mode && pass == 2) Log("\n");
	}

	if (pos >= size || list_cell_mode)
		return 0;

	/* Skip NAMn (4 bytes for marker, 4 bytes for Length),
	 * We also skip the length itself.
	 */
	if (strncmp("NAM", r + pos, 3) == 0) {
		nsize = (int) r[pos+4];
		pos += 8 + nsize;
	}

	/* FRMR record.
	 */

	while (pos < size)  {
		if (strncmp("FRMR", r + pos, 4) == 0 && pass == 2) {
			total_objects++;
			pos += ProcessFRMRData(r + pos, size);
		} else {
			pos += 4;
		}
	}

	return 0;
}

/*******************************************************************
** 3. ProcessFRMRData():
** 	Processes a FRMR subrecord: Each object stored in a CELL record is
** 	a FRMR subrecord.
*******************************************************************
** FRMR (4 bytes) + Length (4 bytes) + Index (4 bytes)
** NAME (4 bytes) + Length (4 bytes) + NameData (length bytes) +
**                + Scale (4) + Scale Length (4) + ScaleData (4)
** FLTV: LockInfo (optional 12 bytes).
** KNAM: KeyInfo (optional 4 KNAM + 4 length + length bytes name).
** TNAM: Trap Info (optional 4 TNAM + 4 length + length bytes name).
** DATA (4 bytes) + Length (4 bytes) + XYZ Data (24 bytes).
********************************************************************/

int ProcessFRMRData(char *r, int record_size)
{
	char name[512];
	int  pos = 0,
	     nsize;
	float xpos, ypos, zpos,  // NB: We only really need xpos & ypos!
	      xrot, yrot, zrot;

	/*
	 * Skip the FRMR and index data (12 bytes) and NAME marker (4).
	 */

	pos += 12;

	if (strncmp("NAME", r + pos, 4) != 0) {
		sprintf(log_message, "Unknown field in FRMR entry: %c%c%c%c\n",
			r[pos], r[pos+1], r[pos+2], r[pos+3]);
		Log(log_message);
		pos = record_size;
		return pos;
	}
	pos += 4;

	nsize = (int) r[pos];
	strncpy(name, r + 4 + pos, nsize);
	pos += 4 + nsize;

	if (verbose_mode) {
		sprintf(log_message, "\tObject: \"%s\"\t", name);
		Log(log_message);
	}

	/************************************************************
	 * Many optional object fields may now follow. We're
	 * only looking for the DATA field (contains the
	 * (xyz co-ords of the object). All other fields are
	 * skipped. Some objects, strangely (e.g. a couple in
	 * "Siege at Firemoth"), don't contain co-ordinates; if FRMR
	 * is encountered then this function returns, so the calling
	 * function can process the next object.
	 * Any unknown fields are spat out as warnings and ignored.
	 ***********************************************************/

	while (strncmp("DATA", r + pos, 4) != 0 && pos < record_size-1) {
		if (strncmp("FLTV", r + pos, 4) == 0 ||
		    strncmp("KNAM", r + pos, 4) == 0 ||
		    strncmp("XSCL", r + pos, 4) == 0 ||
		    strncmp("DNAM", r + pos, 4) == 0 ||
		    strncmp("ANAM", r + pos, 4) == 0 ||
		    strncmp("INTV", r + pos, 4) == 0 ||
		    strncmp("NAM9", r + pos, 4) == 0 ||
		    strncmp("INDX", r + pos, 4) == 0 ||
		    strncmp("TNAM", r + pos, 4) == 0 ||
		    strncmp("DELE", r + pos, 4) == 0 ||
		    strncmp("XCHG", r + pos, 4) == 0 ||
		    strncmp("CNAM", r + pos, 4) == 0) {
			nsize = (int) r[pos+4];
			pos += 8 + nsize;
		} else if (strncmp("DODT", r + pos, 4) == 0 && cell.type == INTERIOR) {
			ProcessDODTData(r + pos, record_size);
			nsize = (int) r[pos+4];
			pos += 8 + nsize;
		} else if (strncmp("DODT", r + pos, 4) == 0 && cell.type == EXTERIOR) {
			nsize = (int) r[pos+4];
			pos += 8 + nsize;
		} else if (strncmp("FRMR", r + pos, 4) == 0) {
			return pos;
		} else {
			if (verbose_mode) {
				sprintf(log_message, "[Warning: <Unknown FRMR field: %c%c%c%c>]!\n",
					r[pos], r[pos+1], r[pos+2], r[pos+3]);
				Log(log_message);
			}
			nsize = bytes_to_int(r[pos+4], r[pos+5], r[pos+6], r[pos+7]);
			pos += 8 + nsize;
		}
	}

	/* What we really wanted all along; the DATA field.
	 * DATA (4 bytes), length (4bytes), X, Y, Z (4 bytes each) and
	 * rotX, rotY, rotZ (4 bytes each) - all 32-bit floats.
	 */

	if (strncmp("DATA", r + pos, 4) == 0) {
		pos += 8;

		memcpy(&xpos, r+pos,   4);
		memcpy(&ypos, r+pos +  4, 4);
		memcpy(&zpos, r+pos +  8, 4);
		memcpy(&xrot, r+pos + 12, 4);
		memcpy(&yrot, r+pos + 16, 4);
		memcpy(&zrot, r+pos + 20, 4);

		if (verbose_mode) {
			sprintf(log_message, "(%f, %f) ", xpos, ypos);
			Log(log_message);
		}

		if (cell.current_x != cell.new_x
		 || cell.current_y != cell.new_y) {
			if (!verbose_mode) {
				sprintf(log_message, "\tObject \"%s\"\t(%f, %f) ",
					name, xpos, ypos);
				Log(log_message);
			}
			xpos += (float) (8192 * (cell.new_x - cell.current_x));
			memcpy(r + pos, &xpos, 4);
			memcpy(&xpos, r + pos, 4);
			ypos += (float) (8192 * (cell.new_y - cell.current_y));
			memcpy(r + pos + 4, &ypos, 4);
			memcpy(&ypos, r + pos +  4, 4);
			sprintf(log_message, " - MODIFIED to (%f, %f)\n", xpos, ypos);
			Log(log_message);
			total_objects_changed++;
		} else if (verbose_mode)
			Log("\n");

		//printf("(%8.3f, %8.3f, %8.3f) (%6.4f, %6.4f, %6.4f)",
		//	xpos, ypos, zpos, xrot, yrot, zrot);
		pos += 24;
	}


	return pos;
}

/************************************************************************
** 4. ProcessDODTData(): Process a DODT teleporting door record.
************************************************************************
** First do an entire scan for a DNAM field - this only exists for
** Doors pointing to Interior locations, which we may ignore.
** Then look for the DODT field which contains the XYZ teleport
** co-ordinates - these values are checked against the temp cell change
** list file and if it points to a location that is to be moved, the
** door's destination X and Y co-ordinates are modified accordingly.
**
** DODT (4 bytes) + Length (4 bytes) + XYZ Data (24 bytes).
***********************************************************************/

int ProcessDODTData(char *r, int record_size)
{
	int  pos = 0,
	     dodt_pos,
	     nsize;
	int  int_x,    		// World Cell (X,Y) that door teleports to.
	     int_y,
	     trans_x = 0,
	     trans_y = 0;
	float xpos, ypos, zpos,  // NB: We only really need xpos & ypos!
	      xrot, yrot, zrot;

	if (verbose_mode) {
		//sprintf(log_message, "\tDoor: \"%s\"\t", name);
		Log("\tInterior Door: ");
	}

	dodt_pos = pos;

	/************************************************************
	 * Many optional object fields may now follow. We scan for
	 * the DATA field or end of record, but our primary concern
	 * is whether a DNAM field exists. If it does, then it's an
	 * Interior->Interior cell and we don't need to worry about it.
	 * Any unknown fields are spat out as warnings and ignored.
	 ***********************************************************/

	while (strncmp("DATA", r + pos, 4) != 0 && pos < record_size-1) {
		if (strncmp("DNAM", r + pos, 4) == 0) {
			if (verbose_mode) {
				Log("[Interior->Interior Door Teleport - ignoring.] ");
			}
			return pos;
		} else if (strncmp("CNAM", r + pos, 4) == 0 ||
			strncmp("DODT", r + pos, 4) == 0) {
				nsize = (int) r[pos+4];
				pos += 8 + nsize;
		} else {
			if (verbose_mode) {
				sprintf(log_message, "[Warning: <Unknown DODT field: %c%c%c%c>]!\n",
					r[pos], r[pos+1], r[pos+2], r[pos+3]);
				Log(log_message);
			}
			nsize = bytes_to_int(r[pos+4], r[pos+5], r[pos+6], r[pos+7]);
			pos += 8 + nsize;
		}
	}

	/* Now we know it's not an Interior->Interior door, move the pointer back to
	 * where the DODT field started and extract the destination co-ordinates.
	 * DODT (4 bytes), length (4bytes), teleport X, Y, Z (4 bytes each) and
	 * rotX, rotY, rotZ (4 bytes each) - all 32-bit floats.
	 */

	pos = dodt_pos;

	if (strncmp("DODT", r + pos, 4) == 0) {
		pos += 8;

		memcpy(&xpos, r+pos,   4);
		memcpy(&ypos, r+pos +  4, 4);
		memcpy(&zpos, r+pos +  8, 4);
		memcpy(&xrot, r+pos + 12, 4);
		memcpy(&yrot, r+pos + 16, 4);
		memcpy(&zrot, r+pos + 20, 4);

		int_x = (int) (xpos / 8192);
		int_y = (int) (ypos / 8192);

		/* Negative co-ordinates need to be decremented by 1.
		 * e.g. (-6000, -12000) exists in cell (-1, -2) not (-0, -1) !
		 */

		if (xpos < 0) {
			int_x -= 1;
		}
		if (ypos < 0) {
			int_y -= 1;
		}

		if (verbose_mode) {
			sprintf(log_message, "Door teleports to (%f, %f): cell [%d, %d] ", xpos, ypos, int_x, int_y);
			Log(log_message);
		}

		if (CheckCellChangeList(int_x, int_y, &trans_x, &trans_y)) {
			if (!verbose_mode) {
				sprintf(log_message, "\tDoor teleport\t(%f, %f) ",
					xpos, ypos);
				Log(log_message);
			}
			xpos += (float) (8192 * (trans_x));
			memcpy(r + pos, &xpos, 4);
			memcpy(&xpos, r + pos, 4);
			ypos += (float) (8192 * (trans_y));
			memcpy(r + pos + 4, &ypos, 4);
			memcpy(&ypos, r + pos +  4, 4);
			sprintf(log_message, " - MODIFIED to (%f, %f)\n", xpos, ypos);
			cell.save = FALSE;
			cell.copy = TRUE;
			Log(log_message);
			total_objects_changed++;
		} else if (verbose_mode)
			Log("\n");

		//printf("(%8.3f, %8.3f, %8.3f) (%6.4f, %6.4f, %6.4f)",
		//	xpos, ypos, zpos, xrot, yrot, zrot);
		pos += 24;
	}

	return pos;
}


/*****************************************************************
** 5. ProcessLandData(): Process a LAND record.
*****************************************************************
** LAND (4 bytes) + Length (4 bytes) + X (4 bytes) + Y (4 bytes).
****************************************************************/

int ProcessLANDData(char *r, int size)
{
	/*********************************************************
	 * Get the (hopefully) INTV  header and X, Y co-ordinates.
	 ********************************************************/
	if (strncmp("INTV", r, 4) == 0) {
		cell.current_x = bytes_to_int(r[8], r[9], r[10], r[11]);
		cell.current_y = bytes_to_int(r[12], r[13], r[14], r[15]);
		//cell.current_x = (int) r[8] ;
		//cell.current_y = (int) r[12];
		cell.new_x = cell.current_x;
		cell.new_y = cell.current_y;
		if (verbose_mode) {
			sprintf(log_message, "LAND co-ords: (%d, %d) ",
				cell.current_x, cell.current_y);
			Log(log_message);
		}
	} else {
		sprintf(log_message, "WARNING: I couldn't find the INTV header (got [%c%c%c%c] - ignoring record.\n",
			r[0], r[1], r[2], r[3]);
		Log(log_message);
		return 1;
	}
	if (cell.region_name[0] != '\0' || cell.name[0] == '\0')
		cell.type = EXTERIOR;

	ReadRules(TESFAITH_RULES_FILE, NULL);
	if (cell.current_x  != cell.new_x || cell.current_y != cell.new_y) {
		memcpy(&r[8], &cell.new_x, 4);
		memcpy(&r[12], &cell.new_y, 4);
		total_records_changed++;
		if (!verbose_mode) {
			sprintf(log_message, "LAND co-ords: (%d, %d), ",
				cell.current_x, cell.current_y);
			Log(log_message);
		}
		sprintf(log_message, " - MODIFIED to (%d, %d)\n",
			bytes_to_int(r[8], r[9], r[10], r[11]),
			bytes_to_int(r[12], r[13], r[14], r[15]));
		Log(log_message);
		cell.region_name[0] = '\0';
	} else if (verbose_mode)
		Log("\n");

	return 0;
}

/**************************************
** 6. Process a SCPT (script) record.
**************************************
** SCPT (4 bytes) + Length (12 bytes).
**************************************/

int ProcessSCPTData(char *r, int *record_size)
{
	int pos = 0,
	    pos_sctx = 0;
	int nsize;

	int ss_record_size = 0;

	char sname[128],
	     line[1024],
	     lower_line[1024];

	FILE *fw, // For writing temporary script text to a file.
	     *fp; // For reading temporary script text from a file.

	/*********************************************************
	 * Get the (hopefully) SCTX header.
	 ********************************************************/

	pos = 16;

	while (strncmp("SCTX", r + pos, 4) != 0 && pos < *record_size-1) {
		if (strncmp("SCHD", r + pos, 4) == 0) {
			nsize = (int) r[pos+4];
			strncpy(sname, &r[pos+8], nsize);
			if (verbose_mode) {
				sprintf(log_message, " - script called \"%s\"", sname);
				Log(log_message);
			}
			pos += 8 + nsize;
		} else if (strncmp("SCVR", r + pos, 4) == 0 ||
		    	strncmp("SCDT", r + pos, 4) == 0) {
				nsize = bytes_to_int(r[pos+4], r[pos+5], r[pos+6], r[pos+7]);
				pos += 8 + nsize;
		} else {
			if (verbose_mode) {
				sprintf(log_message, "[Warning: <Unknown SCDT field: %c%c%c%c>]!\n",
					r[pos], r[pos+1], r[pos+2], r[pos+3]);
				Log(log_message);
			}
			nsize = bytes_to_int(r[pos+4], r[pos+5], r[pos+6], r[pos+7]);
			pos += 8 + nsize;
		}
	}

	if (strncmp("SCTX", r + pos, 4) == 0) {
		nsize = bytes_to_int(r[pos+4], r[pos+5], r[pos+6], r[pos+7]);
		pos_sctx = pos;
		pos += 8;

		if ((fw = fopen(TESFAITH_TMP_SCRIPT_OUT_FILE, "wb")) == 0) {
			fprintf(stderr, "Unable to create temporary script output file for writing, %s: %s\n",
				TESFAITH_TMP_SCRIPT_OUT_FILE, strerror(errno));
			exit(1);
		}

		fwrite(r+pos, 1, nsize, fw);
		fclose(fw);

		if ((fp = fopen(TESFAITH_TMP_SCRIPT_OUT_FILE, "r")) == 0) {
			fprintf(stderr, "Unable to create temporary script output file for reading, %s: %s\n",
				TESFAITH_TMP_SCRIPT_OUT_FILE, strerror(errno));
			exit(1);
		}


		if ((fw = fopen(TESFAITH_TMP_NEW_SCRIPT_OUT_FILE, "w")) == 0) {
			fprintf(stderr, "Unable to open new temporary script output file for writing, %s: %s\n",
				TESFAITH_TMP_NEW_SCRIPT_OUT_FILE, strerror(errno));
			exit(1);
		}

		while (fgets(line, 1024, fp) != NULL) {
			strcpy(lower_line, line);
			strlwr(lower_line);
			if (strstr(lower_line, "position ") > 0 ||
			    strstr(lower_line, "placeitem ") > 0 ||
			    strstr(lower_line, "aifollow ") > 0 ||
			    strstr(lower_line, "aiescort ") > 0 ||
			    strstr(lower_line, "aitravel ") > 0) {
				if (ModifyScriptLine(line) > 0 ) {
					if (!verbose_mode) {
						sprintf(log_message, " - for script called \"%s\"\n", sname);
						Log(log_message);
					}
					total_scripts_changed++;
					ss_record_size += strlen(line) - strlen(lower_line);
					cell.save = FALSE;
					cell.copy = TRUE;
				}
			}
			fprintf(fw, "%s", line);
		}

		fclose(fp);
		fclose(fw);

		if (cell.copy == TRUE) {
			*record_size += ss_record_size;

			/* We've cheated the following bit, originally mallocing an extra 2K
			 * to the record, r, in main() to cope with a compiler option issue.
			 */

			/*if (ss_record_size > 0) {
				if (realloc(r, *record_size) == NULL) {
					fprintf(stderr, "Unable to reallocate %d bytes of memory to store new script: %s\n",
						*record_size, strerror(errno));
					exit(1);
				}
			}*/

			if ((fp = fopen(TESFAITH_TMP_NEW_SCRIPT_OUT_FILE, "rb")) == 0) {
				fprintf(stderr, "Unable to create new temporary script output file for reading, %s: %s\n",
					TESFAITH_TMP_NEW_SCRIPT_OUT_FILE, strerror(errno));
				exit(1);
			}

			/* Size of new SCTX script text field. */

			nsize += ss_record_size;

			/* Read new script from temp file back in to the memory record. */

			if (fread(&r[pos_sctx + 8], 1, nsize, fp) != nsize) {
				fprintf(stderr, "Unable to read back %d bytes from modified script file: %s: %s\n",
					nsize, TESFAITH_TMP_NEW_SCRIPT_OUT_FILE, strerror(errno));
				exit(1);
			}
			fclose(fp);

			/* Modify size of SCTX field. */

			memcpy(&r[pos_sctx+4], &nsize, 4);

			/* Modify overall size of SCPT record. */

			nsize = bytes_to_int(r[4], r[5], r[6], r[7]) + ss_record_size;
			memcpy(&r[4], &nsize, 4);
		}

	}

	if (verbose_mode)
		Log("\n");

	return 0;
}


/**********************************************************************
** 7. ProcessINFOData(): Process a INFO (dialogue information) record.
**********************************************************************
** INFO (4 bytes) + Length (12 bytes).
*********************************************************************/

int ProcessINFOData(char *r, int *record_size)
{
	int pos = 0,
	    pos_bnam = 0;
	int nsize;

	int ii_record_size = 0;

	char name[128],
	     line[1024],
	     lower_line[1024];

	FILE *fw, // For writing temporary script text to a file.
	     *fp; // For reading temporary script text from a file.


	/*********************************************************
	 * Get the (hopefully) BNAM header.
	 ********************************************************/

	pos = 16;

	while (strncmp("BNAM", r + pos, 4) != 0 && pos < *record_size-1) {
		if (strncmp("ONAM", r + pos, 4) == 0) {
			nsize = bytes_to_int(r[pos+4], r[pos+5], r[pos+6], r[pos+7]);
			strncpy(name, &r[pos+8], nsize);
			if (verbose_mode) {
				sprintf(log_message, " - text for ID \"%s\"", name);
				Log(log_message);
			}
			pos += 8 + nsize;
		} else if (strncmp("PNAM", r + pos, 4) == 0 ||
		    	strncmp("NAME", r + pos, 4) == 0 ||
		    	strncmp("NNAM", r + pos, 4) == 0 ||
		    	strncmp("INAM", r + pos, 4) == 0 ||
		    	strncmp("FNAM", r + pos, 4) == 0 ||
		    	strncmp("DNAM", r + pos, 4) == 0 ||
		    	strncmp("ANAM", r + pos, 4) == 0 ||
		    	strncmp("RNAM", r + pos, 4) == 0 ||
		    	strncmp("CNAM", r + pos, 4) == 0 ||
		    	strncmp("SNAM", r + pos, 4) == 0 ||
		    	strncmp("FLTV", r + pos, 4) == 0 ||
		    	strncmp("SCTV", r + pos, 4) == 0 ||
		    	strncmp("SCVR", r + pos, 4) == 0 ||
		    	strncmp("QSTF", r + pos, 4) == 0 ||
		    	strncmp("QSTN", r + pos, 4) == 0 ||
		    	strncmp("QSTR", r + pos, 4) == 0 ||
		    	strncmp("DELE", r + pos, 4) == 0 ||
		    	strncmp("INTV", r + pos, 4) == 0 ||
		    	strncmp("DATA", r + pos, 4) == 0) {
				nsize = bytes_to_int(r[pos+4], r[pos+5], r[pos+6], r[pos+7]);
				pos += 8 + nsize;
		} else {
			if (verbose_mode) {
				sprintf(log_message, "[Warning: <Unknown INFO field: %c%c%c%c>]!\n",
					r[pos], r[pos+1], r[pos+2], r[pos+3]);
				Log(log_message);
			}
			nsize = bytes_to_int(r[pos+4], r[pos+5], r[pos+6], r[pos+7]);
			pos += 8 + nsize;
		}
	}

	if (strncmp("BNAM", r + pos, 4) == 0) {
		nsize = bytes_to_int(r[pos+4], r[pos+5], r[pos+6], r[pos+7]);
		pos_bnam = pos;
		pos += 8;

		if ((fw = fopen(TESFAITH_TMP_SCRIPT_OUT_FILE, "wb")) == 0) {
			fprintf(stderr, "Unable to create temporary script output file for writing, %s: %s\n",
				TESFAITH_TMP_SCRIPT_OUT_FILE, strerror(errno));
			exit(1);
		}

		fwrite(r+pos, 1, nsize, fw);
		fclose(fw);

		if ((fp = fopen(TESFAITH_TMP_SCRIPT_OUT_FILE, "r")) == 0) {
			fprintf(stderr, "Unable to create temporary script output file for reading, %s: %s\n",
				TESFAITH_TMP_SCRIPT_OUT_FILE, strerror(errno));
			exit(1);
		}


		if ((fw = fopen(TESFAITH_TMP_NEW_SCRIPT_OUT_FILE, "w")) == 0) {
			fprintf(stderr, "Unable to open new temporary script output file for writing, %s: %s\n",
				TESFAITH_TMP_NEW_SCRIPT_OUT_FILE, strerror(errno));
			exit(1);
		}

		while (fgets(line, 1024, fp) != NULL) {
			strcpy(lower_line, line);
			strlwr(lower_line);
			if (strstr(lower_line, "position ") > 0 ||
			    strstr(lower_line, "placeitem ") > 0 ||
			    strstr(lower_line, "aifollow ") > 0 ||
			    strstr(lower_line, "aiescort ") > 0 ||
			    strstr(lower_line, "aitravel ") > 0) {
				if (ModifyScriptLine(line) > 0 ) {
					if (!verbose_mode) {
						sprintf(log_message, " - changed Dialog command text for \"%s\"\n", name);
						Log(log_message);
					}
					total_dialogs_changed++;
					ii_record_size += strlen(line) - strlen(lower_line);
					cell.save = FALSE;
					cell.copy = TRUE;
				}
			}
			fprintf(fw, line);
		}

		fclose(fp);
		fclose(fw);

		if (cell.copy == TRUE) {
			*record_size += ii_record_size;

			/* We've cheated the following bit, originally mallocing an extra 2K
			 * to the record, r, in main() to cope with a compiler option issue.
			 */

			/*if (ii_record_size > 0) {
				if (realloc(r, *record_size + 1) == NULL) {
					fprintf(stderr, "Unable to reallocate %d bytes of memory to store new dialog script: %s\n",
						*record_size, strerror(errno));
					exit(1);
				}
			}*/

			if ((fp = fopen(TESFAITH_TMP_NEW_SCRIPT_OUT_FILE, "rb")) == 0) {
				fprintf(stderr, "Unable to create new temporary script output file for reading, %s: %s\n",
					TESFAITH_TMP_NEW_SCRIPT_OUT_FILE, strerror(errno));
				exit(1);
			}

			/* Size of new BNAM Info Dialog Command text field. */

			nsize += ii_record_size;

			/* Read new script from temp file back in to the memory record. */

			fread(r + pos_bnam + 8, 1, nsize, fp);
			fclose(fp);

			/* Modify size of BNAM field. */

			memcpy(&r[pos_bnam+4], &nsize, 4);

			/* Modify overall size of INFO record. */

			nsize = bytes_to_int(r[4], r[5], r[6], r[7]) + ii_record_size;
			memcpy(&r[4], &nsize, 4);
		}

	}

	if (verbose_mode)
		Log("\n");

	return 0;
}

/**********************************************************************
** 8. ModifyScriptLine():
**********************************************************************
** Check whether a Script Command that uses co-ordinates needs
** updating to new co-ordinates and modify the original
** line accordingly.
**
** Returns 0 if no change needed to be made or 1 if line was modified.
*********************************************************************/

int ModifyScriptLine(char *line)
{
	int idx = 0,
	    num = 0,
	    follow = 0,		/* For AIFollow or AIEscort commands: 0,0 co-ords must not be processed. */
	    ignorenums = 0;     /* For AIFollow,AIEscort and PlaceItem: Ignore the first few numbers. */

	int icommars = 0;
	int not_space = 0;

	int int_x, int_y, trans_x, trans_y;

	float fx, fy, new_fx, new_fy;

	char coord[3][32],
	     *p[4],
	     *s,
	     lower_line[1024],
	     new_line[1024];

	strcpy(lower_line, line);
	strlwr(lower_line);

	if ((s = strstr(lower_line, "position ")) == NULL &&
	    (s = strstr(lower_line, "aitravel ")) == NULL) {
		if ((s = strstr(lower_line, "placeitem ")) != NULL) {
			ignorenums = 1;
		} else if ((s = strstr(lower_line, "aifollow ")) != NULL
		    	   || (s = strstr(lower_line, "aiescort ")) != NULL) {
			follow = 1;
			ignorenums = 2;
		}
	}

	/*
	 * If the pointer isn't equal to the beginning of the line and the preceding char is a alphanumeric,
	 * then abort - this is another command (e.g. don't confuse position with moddisposition).
	 */

	if (s != lower_line && isalnum(s[-1])) {
		return 0;
	}

	for (num = 0; num < 2; num++) {
		if (ignorenums > 0 && num == 0) {
			while(!isspace(s[0]) && s[0] != '\0') {
				*s++;
			}
			while(isspace(s[0]) && s[0] != '\0') {
				*s++;
			}
		} else {
			while(!isdigit(s[0]) && s[0] != '-' && s[0] != '.' && s[0] != '\0') {
				*s++;
			}
		}
		if (s[0] == '\0')
			return 0;

		if (ignorenums > 0 && num == 0) {
			while (ignorenums > 0) {
				icommars = 0;
				not_space = 0;
				ignorenums--;

				while(s[0] != '\0' && ((icommars > 0 && icommars < 2) || (s[0] != ' ' 
				   && s[0] != ',' && not_space == 0))) {
					if (s[0] == '\"') icommars++;
					if (isspace(s[0]) || s[0] == ',') not_space++;
					*s++;
				}
				while(!isdigit(s[0]) && s[0] != '-' && s[0] != '.' && s[0] != '\0') {
					*s++;
				}
			}
		}

		/* Record exactly where this number begins. */

		p[2*num] = s;
		idx = 0;
		while ((s[0] == '-' || isdigit(s[0]) || s[0] == '.') && s[0] != '\0') {
			coord[num][idx++] = s[0];
			*s++;
		}

		/* Record where we are between the two numbers (so we know what delimiters there are). */

		p[(2*num)+1] = s;
		coord[num][idx] = '\0';
		//printf("Number %d is >>%s<<\n", num, coord[num]);
	}

	/*
	 * Convert the extracted numbers in to floating point numbers and get the integer cell values.
	 */

	fx = atof(coord[0]);
	fy = atof(coord[1]);

	int_x = (int) (fx / 8192);
	int_y = (int) (fy / 8192);

	/* Negative co-ordinates need to be decremented by 1.
	 * e.g. (-6000, -12000) exists in cell (-1, -2) not (-0, -1) !
	 */

	if (fx < 0) {
		int_x -= 1;
	}
	if (fy < 0) {
		int_y -= 1;
	}

	/*
	 * If this command is an AIFollow or AIEscort command and the X,Y co-ords are both 0,
	 * do not make any changes.
	 */

	if (follow && fx == 0 && fy == 0)	{
		//printf("Ignoring command %s -->it's a follow and the co-ords are 0.\n", line);
		return 0;
	}

	/*
	 * If CheckCellChangeList() returns true then the co-ordinates pointing to this cell
	 * should be modified, so calculate the offset and generate a modified version of the
	 * Script command.
	 */

	if (CheckCellChangeList(int_x, int_y, &trans_x, &trans_y)) {
		new_fx = fx + (8192 * trans_x);
		new_fy = fy + (8192 * trans_y);

		sprintf(log_message, "\n\tMODIFIED script usage of co-ords (%f, %f) to (%f,%f)",
			fx, fy, new_fx, new_fy);
		Log(log_message);

		/* Reformulate the new line. Firstly everything up to the beginning of the first number. */

		strncpy(new_line, line, p[0] - lower_line);
		new_line[p[0] - lower_line] = '\0';

		/* Then the first number, then everything between the two numbers (delimiters etc.) */

		sprintf(new_line, "%s%f", new_line, new_fx);
		strncat(new_line, line + (p[1] - lower_line), p[2] - p[1]);

		/* Then the second number, followed by the remainder of the original line. */

		sprintf(new_line, "%s%f", new_line, new_fy);
		strcat(new_line, line + (p[3] - lower_line));

		/* Copy the new line over the original line. */

		strcpy(line, new_line);
		return 1;
	}

	return 0;
}

/**************************************************************************
** 9. ReadRulesCLArgs():
**
** Called if no command line arguments are specified, this routine will
** scan the Rules configuration file for a "tesfaith " line and populate
** the *argv[] command line array using these parameters.
*************************************************************************/

void ReadRulesCLArgs(char *filename, int *argc, char *argv[])
{
	char line[1024],
	     lower_line[1024],
	     sarg[1024],
	     *s;

	int p = 0,
	    pn = 0,
	    inquote = 0;

	FILE *fp;

	/****************************************
	 ** Open the TESfaith configuration file.
	 ***************************************/

	if ((fp = fopen(filename, "r")) == NULL) {
		fprintf(stderr, "Unable to open your Cell Transposition "
			"rules file (%s): %s\n",
			filename, strerror(errno));
		exit(1);
	}

	while (fgets(line, 1024, fp) != NULL) {
		strcpy(lower_line, line);
		strlwr(lower_line);

		if (strncmp(lower_line, "tesfaith ", 9) == 0) {
			p = 8;
			while (line[p] != '\n' && line[p] != '\r') {
				inquote = 0;
				while (line[p] != '\n' && line[p] != '\r' && line[p] == ' ' && (line[p] != '\"')) p++;
				if (line[p] == '\"') {
					inquote++;
					p++;
				}
				pn = 0;
				while (line[p] != '\n' && line[p] != '\r' && (!isspace(line[p]) || inquote) && line[p] != '\"') {
					sarg[pn++] = line[p++];
				}
				sarg[pn] = '\0';
				if (line[p] == '\"') p++;
				if ((s = malloc(strlen(sarg))) == NULL) {
					printf("Unable to malloc space for command line arguments: %s\n",
						strerror(errno));
				}
				//printf("arg is :%s:\n", sarg);
				strcpy(s, sarg);
				argv[*argc] = s;
				*argc += 1;
			}
			argv[*argc] = NULL;
			fclose(fp);
			return;
		}
	}

	fclose(fp);
}

/**********************************************************************************
** 10. ReadRules():
**
** Read the rules from the TESfaith rules file and process them accordingly.
** cell_record is only populated for INterior Cells (it contains the cell's name).
*********************************************************************************/

void ReadRules(char *filename, char *cell_record)
{
	char line[1024];
	int  rule_no = 0;

	FILE *fp;

	/****************************************
	 ** Open the TESfaith configuration file.
	 ***************************************/

	if ((fp = fopen(filename, "r")) == NULL) {
		fprintf(stderr, "Unable to open your Cell Transposition "
			"rules file (%s): %s\n",
			filename, strerror(errno));
		exit(1);
	}

	/**************************************************************
	 * Read each line from the configuration file.
	 * Lines beginning with # (comments) or spaces are ignored.
	 *************************************************************/

	while (fgets(line, 1024, fp) != NULL) {
		if (!isalpha(line[0]))
			continue;

		/*
		 * Translate Region.
	 	 * TR, "Region ID Name", X-offset, Y-offset.
		 */

		if (ParseRule(line) != 0) {
			exit(1);
		}
		rule_no++;

		if (cell.type == EXTERIOR) {
			/**************************************
			 ** Rule 1: 'TA' - Transpose All Cells.
			 *************************************/
			if (strncmp("TA", line, 2) == 0) {
				if (cell.size == 0) {
					sprintf(log_message, "Rule %d: All external cells will be transposed by (%d, %d) cells.\n",
						rule_no, atoi(rule[1]), atoi(rule[2]));
					Log(log_message);
					if (show_rules_mode) printf(log_message);
					continue;
				} else {
					if (pass == 2) Log(" [TA: Transpose All rule matched] ");
					cell.copy = TRUE;
					cell.save = FALSE;
					cell.new_x += atoi(rule[1]);;
					cell.new_y += atoi(rule[2]);;
					fclose(fp);
					return;
				}
			/**************************************
			 ** Rule 2: 'CA' - Copy All Cells.
			 *************************************/
			} else if (strncmp("CA", line, 2) == 0) {
				if (cell.size == 0) {
					sprintf(log_message, "Rule %d: All external cells will be copied by an offset of (%d, %d) cells.\n",
						rule_no, atoi(rule[1]), atoi(rule[2]));
					Log(log_message);
					if (show_rules_mode) printf(log_message);
					continue;
				} else {
					if (pass == 2) Log(" [CA: Transpose All rule matched] ");
					cell.copy = TRUE;
					cell.save = TRUE;
					cell.new_x += atoi(rule[1]);;
					cell.new_y += atoi(rule[2]);;
					fclose(fp);
					return;
				}
			/***********************************
			 ** Rule 3: 'TR' - Transpose Region.
			 **********************************/
			} else if (strncmp("TR", line, 2) == 0) {
				if (cell.size == 0) {
					sprintf(log_message, "Rule %d: All cells in the region (%s) will be transposed by (%d, %d) cells.\n",
						rule_no, rule[1], atoi(rule[2]), atoi(rule[3]));
					Log(log_message);
					if (show_rules_mode) printf(log_message);
					continue;
				} else if (strcmp(cell.region_name, rule[1]) == 0) {
					if (pass == 2) Log(" [TR: Transpose Region rule matched] ");
					cell.copy = TRUE;
					cell.save = FALSE;
					cell.new_x += atoi(rule[2]);;
					cell.new_y += atoi(rule[3]);;
					fclose(fp);
					return;
				}
			/***********************************
			 ** Rule 4: 'CR' - Copy Region.
			 **********************************/
			} else if (strncmp("CR", line, 2) == 0) {
				if (cell.size == 0) {
					sprintf(log_message, "Rule %d: A copy of each cell in the region (%s) "
						"will be placed at an offset of (%d, %d) cells.\n",
						rule_no, rule[1], atoi(rule[2]), atoi(rule[3]));
					Log(log_message);
					if (show_rules_mode) printf(log_message);
					continue;
				} else if (strcmp(cell.region_name, rule[1]) == 0) {
					if (pass == 2) Log(" [CR: Copy Region rule matched] ");
					cell.copy = TRUE;
					cell.save = TRUE;
					cell.new_x += atoi(rule[2]);;
					cell.new_y += atoi(rule[3]);;
					fclose(fp);
					return;
				}
			/***********************************
			 ** Rule 5: 'MC' - Move Cell.
			 **********************************/
			} else if (strncmp("MC", line, 2) == 0) {
				if (cell.size == 0) {
					sprintf(log_message, "Rule %d: Move the cell at (%d, %d) to (%d, %d).\n",
						rule_no, atoi(rule[1]), atoi(rule[2]),
						atoi(rule[3]), atoi(rule[4]));
					Log(log_message);
					if (show_rules_mode) printf(log_message);
					continue;
				} else if (cell.current_x == atoi(rule[1])
					&& cell.current_y == atoi(rule[2])) {
					if (pass == 2) Log(" [MC: Move Cell rule matched] ");
					cell.copy = TRUE;
					cell.save = FALSE;
					cell.new_x = atoi(rule[3]);;
					cell.new_y = atoi(rule[4]);;
					fclose(fp);
					return;
				}
			/*************************************
			 ** Rule 6: 'CE' - Copy Exterior Cell.
			 ************************************/
			} else if (strncmp("CE", line, 2) == 0) {
				if (cell.size == 0) {
					sprintf(log_message, "Rule %d: Make a copy of the cell (%d, %d) at (%d, %d).\n",
						rule_no, atoi(rule[1]), atoi(rule[2]),
						atoi(rule[3]), atoi(rule[4]));
					Log(log_message);
					if (show_rules_mode) printf(log_message);
					continue;
				} else if (cell.current_x == atoi(rule[1])
				   	&& cell.current_y == atoi(rule[2])) {
					if (pass == 2) Log(" [CE: Copy Exterior Cell rule matched] ");
					cell.copy = TRUE;
					cell.save = TRUE;
					cell.new_x = atoi(rule[3]);;
					cell.new_y = atoi(rule[4]);;
					fclose(fp);
					return;
				}
			/*********************************
			 ** Rule 7: 'TC' - Transpose Cell.
			 ********************************/
			} else if (strncmp("TC", line, 2) == 0) {
				if (cell.size == 0) {
					sprintf(log_message, "Rule %d: Move the cell at (%d, %d) by an offset of (%d, %d).\n",
						rule_no, atoi(rule[1]), atoi(rule[2]),
						atoi(rule[3]), atoi(rule[4]));
					Log(log_message);
					if (show_rules_mode) printf(log_message);
					continue;
				} else if (cell.current_x == atoi(rule[1])
				   	&& cell.current_y == atoi(rule[2])) {
					if (pass == 2) Log(" [TC: Transpose Exterior Cell rule matched] ");
					cell.copy = TRUE;
					cell.save = FALSE;
					cell.new_x += atoi(rule[3]);;
					cell.new_y += atoi(rule[4]);;
					fclose(fp);
					return;
				}
			/***********************************
			 ** Rule 8: 'DA' - Delete All Cells.
			 **********************************/
			} else if (strncmp("DA", line, 2) == 0) {
				if (cell.size == 0) {
					sprintf(log_message, "Rule %d: All external cells will be deleted.\n",
						rule_no);
					Log(log_message);
					if (show_rules_mode) printf(log_message);
					continue;
				} else {
					sprintf(log_message, "Exterior Cell at (%d,%d) DELETED [Delete All matched].\n",
						cell.current_x, cell.current_y);
					if (pass == 2) Log(log_message);
					cell.copy = FALSE;
					cell.save = FALSE;
					total_records_changed++;
					fclose(fp);
					return;
				}
			/*****************************************
			 ** Rule 9: 'DR' - Delete Cells in Region.
			 ****************************************/
			} else if (strncmp("DR", line, 2) == 0) {
				if (cell.size == 0) {
					sprintf(log_message, "Rule %d: Any cells within the region (%s) will be deleted.\n",
						rule_no, rule[1]);
					Log(log_message);
					if (show_rules_mode) printf(log_message);
					continue;
				} else if (strcmp(cell.region_name, rule[1]) == 0) {
					sprintf(log_message, "Exterior Cell at (%d,%d) DELETED [Delete Region matched].\n",
						cell.current_x, cell.current_y);
					if (pass == 2) Log(log_message);
					cell.copy = FALSE;
					cell.save = FALSE;
					total_records_changed++;
					fclose(fp);
					return;
				}
			/***************************************
			 ** Rule 10: 'DE' - Delete Exterior Cell.
			 **************************************/
			} else if (strncmp("DE", line, 2) == 0) {
				if (cell.size == 0) {
					sprintf(log_message, "Rule %d: Any cells found at (%d, %d) will be deleted.\n",
						rule_no, atoi(rule[1]), atoi(rule[2]));
					Log(log_message);
					if (show_rules_mode) printf(log_message);
					continue;
				} else if (cell.current_x == atoi(rule[1])
				   	    && cell.current_y == atoi(rule[2])) {
					sprintf(log_message, "Exterior Cell at (%d,%d) DELETED [DE: Delete Exterior Cell matched].\n",
						cell.current_x, cell.current_y);
					if (pass == 2) Log(log_message);
					cell.copy = FALSE;
					cell.save = FALSE;
					total_records_changed++;
					fclose(fp);
					return;
				}
			} else {
				rule_no--;
				//printf("Invalid Rule command: '%s': Ignoring ...\n", rule[0]);
			}
		} else if (cell.type == INTERIOR) {
			/*************************************
			 ** Rule 11: 'CI' - Copy Interior Cell.
			 ************************************/
			if (strncmp("CI", line, 2) == 0) {
				if (cell.size == 0) {
					sprintf(log_message, "Rule %d: If an interior cell called (%s) is found, a copy of it called (%s) will be created.\n",
						rule_no, rule[1],rule[2]);
					Log(log_message);
					if (show_rules_mode) printf(log_message);
					if (strlen(rule[2]) > strlen(rule[1])) {
						sprintf(log_message, "   Actually I cannot do this yet: please make the copy name the same size or shorter than the original cell name: You can rename the copy in TES.\n");
						Log(log_message);
						if (show_rules_mode)
							printf(log_message);
						exit(1);
					}
					continue;
				} else if (strcmp(cell.name, rule[1]) == 0) {
					sprintf(log_message, " [CI: Delete Exterior Cell rule matched] "
						" - making a copy of \"%s\" called \"%s\"\n",
						cell.name, rule[2]);
					if (pass == 2) Log(log_message);
					//printf("COPY Interior Cell matched!\n");
					memcpy(cell_record+8, rule[2], strlen(rule[2])+1);
					cell.copy = TRUE;
					cell.save = TRUE;
					total_records_changed++;
					//fclose(fp);
					//return;
				}
			/***************************************
			 ** Rule 12: 'DI' - Delete Interior Cell.
			 **************************************/
			} else if (strncmp("DI", line, 2) == 0) {
				if (cell.size == 0) {
					sprintf(log_message, "Rule %d: Any interior cells called (%s) will be deleted.\n",
						rule_no, rule[1]);
					Log(log_message);
					if (show_rules_mode) printf(log_message);
					continue;
				} else if (strcmp(cell.name, rule[1]) == 0) {
					if (pass == 2) Log(" [DI: Delete Interior Cell rule matched] ");
					cell.copy = FALSE;
					cell.save = FALSE;
					total_records_changed++;
					fclose(fp);
					return;
				}
			} else {
				//printf("Invalid Rule command: '%s': Ignoring ...\n", rule[0]);
				rule_no--;
			}

		}

	}

	if (fp != NULL) fclose(fp);
}

/****************************************************
** 11. bytes_to_int():
**
** Convert 4-bytes in to an Integer (little endian).
** Most commonly called for converting the size of a
** record or field in to a usable integer.
***************************************************/

int bytes_to_int(char b1, char b2, char b3, char b4)
{
	return (int) ((unsigned char) b1
		+ (256 * (unsigned char) b2)
		+ (65536 * (unsigned char) b3)
		+ (16777216 * (unsigned char) b4));

}

/**************************************************************
** 12. ParseRule(line): Parse the rule string (input parameter,
** read from the RulesFile), storing each commar delimited
** field in to the rule[][] array. Whitespaces are ignored,
** except when a field is quoted (in which case both
** whitespaces and commars are stored as part of the field).
***************************************************************/

int ParseRule(char *line)
{
	int i = 0;
	int j = 0;
	int index = 0;
	int quote = 0;

	/*****************************************
	 * Delete any rule data left in memory ...
	 ****************************************/

	for (index = 0; index < 5; index++) {
		rule[index][0] = '\0';
	}
	index = 0;

	while(line[i] != '\0') {
		/*
		 * If we're on the fifth command (index starts at 0),
		 * then there must be an error with the rule; the longest
		 * rule can only be 5 fields.
		 */
		if (index > 4) {
			fprintf(stderr, "Invalid Rule: You've specified too many parameters for this command!\n");
			return -1;
		}
		/*
		 * Ignore whitespaces, unless we're in a quoted string.
		 */

		if (line[i] == ' ' && quote == 0) {
			i++;
			continue;
		}

		/*
		 * Toggle the quoted flag; then we know when to record
		 * whitespaces and commars (which are normally ignored).
		 */

		if (line[i] == '\"') {
			if (quote > 0) {
				i++;
				quote = 0;
				continue;
			} else {
				i++;
				quote++;
				continue;
			}
		}
		if ((line[i] != ',' && line[i] != '\r' && line[i] != '\n')
		 || (line[i] == ',' && quote > 0)) {
			rule[index][j++] = line[i];
		} else {
			rule[index][j] = '\0';
			j = 0;
			index++;
		}
		i++;
	}

	return 0;
}

void Log(char *info)
{
	FILE *fp;

	if ((fp = fopen(TESFAITH_LOG_FILE, "a")) == 0) {
		fprintf(stderr, "Unable to open log file for appending "
			"log data (%s): %s\n",
			TESFAITH_LOG_FILE, strerror(errno));
		return;
	}

	fprintf(fp, info);
	fflush(fp);

	fclose(fp);
}

/***************************************************************************************
** 14. WriteCellChangeList():
** Called on the first pass, this routine dumps a list of all individual 
cells that
** will need changing. The temporary file dumped is only used by 
ProcessDODTData() at
** present, but may be used to determine which script entries need modifying 
in future.
**************************************************************************************/
void WriteCellChangeList(int old_x, int old_y, int trans_x, int trans_y)
{
	FILE *fp;

	if ((fp = fopen(TESFAITH_CHANGE_LIST_FILE, "a")) == NULL) {
		fprintf(stderr, "Unable to open change list file for appending "
			"(%s): %s\n",
			TESFAITH_CHANGE_LIST_FILE, strerror(errno));
		return;
	}

	fprintf(fp, "%d, %d, %d, %d\n", old_x, old_y, trans_x, trans_y);
	fflush(fp);

	fclose(fp);
}

/***************************************************************************************
** 15. CheckCellChangeList():
** This routine is used to determine whether a particular cell will be modified. This is
** useful for the Interior->Exterior door teleport references as well as scripts.
**************************************************************************************/
int CheckCellChangeList(int old_x, int old_y, int *trans_x, int *trans_y)
{
	FILE *fp;
	char line[128];

	int x, y;

	if ((fp = fopen(TESFAITH_CHANGE_LIST_FILE, "r")) == NULL) {
		//fprintf(stderr, "Unable to open change list file for reading "
		//	"(%s): %s\n", TESFAITH_CHANGE_LIST_FILE, strerror(errno));
		return 0;
	}

	while (fgets(line, 128, fp) != NULL) {
		sscanf(line, "%d, %d, %d, %d\n", &x, &y, trans_x, trans_y);

		/* If the x and y values read from the file match the old_x and old_y
		 * values that we've been asked to test, then return true (1). The translation
		 * values have already been read in to trans_x and trans_y above.
		 */

		if (x == old_x && y == old_y) {
			fclose(fp);
			return 1;
		}
	}

	fclose(fp);

	return 0;
}



/************************************************************************
** 16. LogCell():
**
** If '-l' was specified on the command-line then this procedure is
** called for each new cell found. A simple dump is made of the cell name
** and it's location; one line as a '#' comment and another as the raw
** values (easier for someone to turn in to a TESfaith rules file).
**********************************************************************/

void LogCell(void)
{
	FILE *fp;

	if ((fp = fopen(TESFAITH_CELL_LIST_FILE, "a")) == 0) {
		fprintf(stderr, "Unable to open %s for writing a list "
			"of the cells in your ESP file: %s\n",
			TESFAITH_CELL_LIST_FILE, strerror(errno));
		exit(1);
	}

	fprintf(fp, "# (%d,%d)\t\"%s\"\t\"%s\"\r\n",
		cell.current_x, cell.current_y, cell.name, cell.region_name);
	fprintf(fp, "%d, %d\r\n", cell.current_x, cell.current_y);

	fclose(fp);
}
